@@ -1,19 +1,19 @@
-from glob import glob
 from distutils import log
 import distutils.command.sdist as orig
 import os
 import sys
 import io
+import contextlib
 
-from setuptools.compat import PY3
-from setuptools.utils import cs_path_exists
+from setuptools.extern import six
+
+from .py36compat import sdist_add_defaults
 
 import pkg_resources
 
-READMES = 'README', 'README.rst', 'README.txt'
-
 _default_revctrl = list
 
+
 def walk_revctrl(dirname=''):
     """Find all files under revision control"""
     for ep in pkg_resources.iter_entry_points('setuptools.file_finders'):
@@ -21,7 +21,7 @@ def walk_revctrl(dirname=''):
             yield item
 
 
-class sdist(orig.sdist):
+class sdist(sdist_add_defaults, orig.sdist):
     """Smart sdist that finds anything supported by revision control"""
 
     user_options = [
@@ -37,6 +37,9 @@ class sdist(orig.sdist):
 
     negative_opt = {}
 
+    README_EXTENSIONS = ['', '.rst', '.txt', '.md']
+    READMES = tuple('README{0}'.format(ext) for ext in README_EXTENSIONS)
+
     def run(self):
         self.run_command('egg_info')
         ei_cmd = self.get_finalized_command('egg_info')
@@ -63,6 +66,45 @@ class sdist(orig.sdist):
             if data not in dist_files:
                 dist_files.append(data)
 
+    def initialize_options(self):
+        orig.sdist.initialize_options(self)
+
+        self._default_to_gztar()
+
+    def _default_to_gztar(self):
+        # only needed on Python prior to 3.6.
+        if sys.version_info >= (3, 6, 0, 'beta', 1):
+            return
+        self.formats = ['gztar']
+
+    def make_distribution(self):
+        """
+        Workaround for #516
+        """
+        with self._remove_os_link():
+            orig.sdist.make_distribution(self)
+
+    @staticmethod
+    @contextlib.contextmanager
+    def _remove_os_link():
+        """
+        In a context, remove and restore os.link if it exists
+        """
+
+        class NoValue:
+            pass
+
+        orig_val = getattr(os, 'link', NoValue)
+        try:
+            del os.link
+        except Exception:
+            pass
+        try:
+            yield
+        finally:
+            if orig_val is not NoValue:
+                setattr(os, 'link', orig_val)
+
     def __read_template_hack(self):
         # This grody hack closes the template file (MANIFEST.in) if an
         #  exception occurs during read_template.
@@ -70,7 +112,7 @@ class sdist(orig.sdist):
         #  file.
         try:
             orig.sdist.read_template(self)
-        except:
+        except Exception:
             _, _, tb = sys.exc_info()
             tb.tb_next.tb_frame.f_locals['template'].close()
             raise
@@ -86,35 +128,8 @@ class sdist(orig.sdist):
     if has_leaky_handle:
         read_template = __read_template_hack
 
-    def add_defaults(self):
-        standards = [READMES,
-                     self.distribution.script_name]
-        for fn in standards:
-            if isinstance(fn, tuple):
-                alts = fn
-                got_it = 0
-                for fn in alts:
-                    if cs_path_exists(fn):
-                        got_it = 1
-                        self.filelist.append(fn)
-                        break
-
-                if not got_it:
-                    self.warn("standard file not found: should have one of " +
-                              ', '.join(alts))
-            else:
-                if cs_path_exists(fn):
-                    self.filelist.append(fn)
-                else:
-                    self.warn("standard file '%s' not found" % fn)
-
-        optional = ['test/test*.py', 'setup.cfg']
-        for pattern in optional:
-            files = list(filter(cs_path_exists, glob(pattern)))
-            if files:
-                self.filelist.extend(files)
-
-        # getting python files
+    def _add_defaults_python(self):
+        """getting python files"""
         if self.distribution.has_pure_modules():
             build_py = self.get_finalized_command('build_py')
             self.filelist.extend(build_py.get_source_files())
@@ -127,26 +142,23 @@ class sdist(orig.sdist):
                     self.filelist.extend([os.path.join(src_dir, filename)
                                           for filename in filenames])
 
-        if self.distribution.has_ext_modules():
-            build_ext = self.get_finalized_command('build_ext')
-            self.filelist.extend(build_ext.get_source_files())
-
-        if self.distribution.has_c_libraries():
-            build_clib = self.get_finalized_command('build_clib')
-            self.filelist.extend(build_clib.get_source_files())
-
-        if self.distribution.has_scripts():
-            build_scripts = self.get_finalized_command('build_scripts')
-            self.filelist.extend(build_scripts.get_source_files())
+    def _add_defaults_data_files(self):
+        try:
+            if six.PY2:
+                sdist_add_defaults._add_defaults_data_files(self)
+            else:
+                super()._add_defaults_data_files()
+        except TypeError:
+            log.warn("data_files contains unexpected objects")
 
     def check_readme(self):
-        for f in READMES:
+        for f in self.READMES:
             if os.path.exists(f):
                 return
         else:
             self.warn(
                 "standard file not found: should have one of " +
-                ', '.join(READMES)
+                ', '.join(self.READMES)
             )
 
     def make_release_tree(self, base_dir, files):
@@ -178,10 +190,10 @@ class sdist(orig.sdist):
         distribution.
         """
         log.info("reading manifest file '%s'", self.manifest)
-        manifest = open(self.manifest, 'rbU')
+        manifest = open(self.manifest, 'rb')
         for line in manifest:
             # The manifest must contain UTF-8. See #303.
-            if PY3:
+            if six.PY3:
                 try:
                     line = line.decode('UTF-8')
                 except UnicodeDecodeError:
